@@ -8,8 +8,9 @@
 #![allow(clippy::exhaustive_structs)]
 
 use std::{
+    borrow::Borrow,
     collections::{HashMap, HashSet},
-    convert::{TryFrom, TryInto},
+    convert::TryInto,
     sync::{
         atomic::{AtomicU64, Ordering::SeqCst},
         Arc,
@@ -18,20 +19,23 @@ use std::{
 
 use criterion::{criterion_group, criterion_main, Criterion};
 use event::StateEvent;
-use js_int::uint;
+use js_int::{int, uint};
 use maplit::{btreemap, hashmap, hashset};
 use ruma_common::MilliSecondsSinceUnixEpoch;
 use ruma_events::{
     pdu::{EventHash, Pdu, RoomV3Pdu},
     room::{
-        join_rules::JoinRule,
-        member::{MemberEventContent, MembershipState},
+        join_rules::{JoinRule, RoomJoinRulesEventContent},
+        member::{MembershipState, RoomMemberEventContent},
     },
     EventType,
 };
-use ruma_identifiers::{EventId, RoomId, RoomVersionId, UserId};
-use ruma_state_res::{Error, Event, EventMap, Result, StateMap, StateResolution};
-use serde_json::{json, Value as JsonValue};
+use ruma_identifiers::{room_id, user_id, EventId, RoomId, RoomVersionId, UserId};
+use ruma_state_res::{self as state_res, Error, Event, Result, StateMap};
+use serde_json::{
+    json,
+    value::{to_raw_value as to_raw_json_value, RawValue as RawJsonValue},
+};
 
 static SERVER_TIMESTAMP: AtomicU64 = AtomicU64::new(0);
 
@@ -45,8 +49,8 @@ fn lexico_topo_sort(c: &mut Criterion) {
             event_id("p") => hashset![event_id("o")],
         };
         b.iter(|| {
-            let _ = StateResolution::lexicographical_topological_sort(&graph, |id| {
-                Ok((0, MilliSecondsSinceUnixEpoch(uint!(0)), id.clone()))
+            let _ = state_res::lexicographical_topological_sort(&graph, |_id| {
+                Ok((int!(0), MilliSecondsSinceUnixEpoch(uint!(0))))
             });
         })
     });
@@ -60,18 +64,15 @@ fn resolution_shallow_auth_chain(c: &mut Criterion) {
         let (state_at_bob, state_at_charlie, _) = store.set_up();
 
         b.iter(|| {
-            let ev_map: EventMap<Arc<StateEvent>> = store.0.clone();
-            let state_sets = vec![state_at_bob.clone(), state_at_charlie.clone()];
-            let _ = match StateResolution::resolve::<StateEvent, _>(
-                &room_id(),
-                &RoomVersionId::Version6,
-                &state_sets,
+            let ev_map = store.0.clone();
+            let state_sets = [&state_at_bob, &state_at_charlie];
+            let _ = match state_res::resolve(
+                &RoomVersionId::V6,
+                state_sets,
                 state_sets
                     .iter()
                     .map(|map| {
-                        store
-                            .auth_event_ids(&room_id(), &map.values().cloned().collect::<Vec<_>>())
-                            .unwrap()
+                        store.auth_event_ids(room_id(), map.values().cloned().collect()).unwrap()
                     })
                     .collect(),
                 |id| ev_map.get(id).map(Arc::clone),
@@ -101,7 +102,12 @@ fn resolve_deeper_event_set(c: &mut Criterion) {
             inner.get(&event_id("PA")).unwrap(),
         ]
         .iter()
-        .map(|ev| ((ev.kind(), ev.state_key().unwrap()), ev.event_id().clone()))
+        .map(|ev| {
+            (
+                (ev.event_type().to_owned(), ev.state_key().unwrap().to_owned()),
+                ev.event_id().to_owned(),
+            )
+        })
         .collect::<StateMap<_>>();
 
         let state_set_b = [
@@ -114,21 +120,23 @@ fn resolve_deeper_event_set(c: &mut Criterion) {
             inner.get(&event_id("PA")).unwrap(),
         ]
         .iter()
-        .map(|ev| ((ev.kind(), ev.state_key().unwrap()), ev.event_id().clone()))
+        .map(|ev| {
+            (
+                (ev.event_type().to_owned(), ev.state_key().unwrap().to_owned()),
+                ev.event_id().to_owned(),
+            )
+        })
         .collect::<StateMap<_>>();
 
         b.iter(|| {
-            let state_sets = vec![state_set_a.clone(), state_set_b.clone()];
-            let _ = match StateResolution::resolve::<StateEvent, _>(
-                &room_id(),
-                &RoomVersionId::Version6,
-                &state_sets,
+            let state_sets = [&state_set_a, &state_set_b];
+            let _ = match state_res::resolve(
+                &RoomVersionId::V6,
+                state_sets,
                 state_sets
                     .iter()
                     .map(|map| {
-                        store
-                            .auth_event_ids(&room_id(), &map.values().cloned().collect::<Vec<_>>())
-                            .unwrap()
+                        store.auth_event_ids(room_id(), map.values().cloned().collect()).unwrap()
                     })
                     .collect(),
                 |id| inner.get(id).map(Arc::clone),
@@ -154,19 +162,19 @@ criterion_main!(benches);
 //  IMPLEMENTATION DETAILS AHEAD
 //
 /////////////////////////////////////////////////////////////////////*/
-pub struct TestStore<E: Event>(pub HashMap<EventId, Arc<E>>);
+struct TestStore<E: Event>(HashMap<Box<EventId>, Arc<E>>);
 
 #[allow(unused)]
 impl<E: Event> TestStore<E> {
-    pub fn get_event(&self, room_id: &RoomId, event_id: &EventId) -> Result<Arc<E>> {
+    fn get_event(&self, room_id: &RoomId, event_id: &EventId) -> Result<Arc<E>> {
         self.0
             .get(event_id)
             .map(Arc::clone)
-            .ok_or_else(|| Error::NotFound(format!("{} not found", event_id.to_string())))
+            .ok_or_else(|| Error::NotFound(format!("{} not found", event_id)))
     }
 
     /// Returns the events that correspond to the `event_ids` sorted in the same order.
-    pub fn get_events(&self, room_id: &RoomId, event_ids: &[EventId]) -> Result<Vec<Arc<E>>> {
+    fn get_events(&self, room_id: &RoomId, event_ids: &[Box<EventId>]) -> Result<Vec<Arc<E>>> {
         let mut events = vec![];
         for id in event_ids {
             events.push(self.get_event(room_id, id)?);
@@ -175,13 +183,9 @@ impl<E: Event> TestStore<E> {
     }
 
     /// Returns a Vec of the related auth events to the given `event`.
-    pub fn auth_event_ids(
-        &self,
-        room_id: &RoomId,
-        event_ids: &[EventId],
-    ) -> Result<HashSet<EventId>> {
+    fn auth_event_ids(&self, room_id: &RoomId, event_ids: Vec<E::Id>) -> Result<HashSet<E::Id>> {
         let mut result = HashSet::new();
-        let mut stack = event_ids.to_vec();
+        let mut stack = event_ids;
 
         // DFS for auth event chain
         while !stack.is_empty() {
@@ -192,25 +196,21 @@ impl<E: Event> TestStore<E> {
 
             result.insert(ev_id.clone());
 
-            let event = self.get_event(room_id, &ev_id)?;
+            let event = self.get_event(room_id, ev_id.borrow())?;
 
-            stack.extend(event.auth_events().clone());
+            stack.extend(event.auth_events().map(ToOwned::to_owned));
         }
 
         Ok(result)
     }
 
-    /// Returns a Vec<EventId> representing the difference in auth chains of the given `events`.
-    pub fn auth_chain_diff(
-        &self,
-        room_id: &RoomId,
-        event_ids: Vec<Vec<EventId>>,
-    ) -> Result<Vec<EventId>> {
+    /// Returns a vector representing the difference in auth chains of the given `events`.
+    fn auth_chain_diff(&self, room_id: &RoomId, event_ids: Vec<Vec<E::Id>>) -> Result<Vec<E::Id>> {
         let mut auth_chain_sets = vec![];
         for ids in event_ids {
             // TODO state store `auth_event_ids` returns self in the event ids list
             // when an event returns `auth_event_ids` self is not contained
-            let chain = self.auth_event_ids(room_id, &ids)?.into_iter().collect::<HashSet<_>>();
+            let chain = self.auth_event_ids(room_id, ids)?.into_iter().collect::<HashSet<_>>();
             auth_chain_sets.push(chain);
         }
 
@@ -218,9 +218,13 @@ impl<E: Event> TestStore<E> {
             let common = auth_chain_sets
                 .iter()
                 .skip(1)
-                .fold(first, |a, b| a.intersection(b).cloned().collect::<HashSet<EventId>>());
+                .fold(first, |a, b| a.intersection(b).cloned().collect::<HashSet<_>>());
 
-            Ok(auth_chain_sets.into_iter().flatten().filter(|id| !common.contains(id)).collect())
+            Ok(auth_chain_sets
+                .into_iter()
+                .flatten()
+                .filter(|id| !common.contains(id.borrow()))
+                .collect())
         } else {
             Ok(vec![])
         }
@@ -228,17 +232,20 @@ impl<E: Event> TestStore<E> {
 }
 
 impl TestStore<StateEvent> {
-    pub fn set_up(&mut self) -> (StateMap<EventId>, StateMap<EventId>, StateMap<EventId>) {
-        let create_event = to_pdu_event::<EventId>(
+    #[allow(clippy::type_complexity)]
+    fn set_up(
+        &mut self,
+    ) -> (StateMap<Box<EventId>>, StateMap<Box<EventId>>, StateMap<Box<EventId>>) {
+        let create_event = to_pdu_event::<&EventId>(
             "CREATE",
             alice(),
             EventType::RoomCreate,
             Some(""),
-            json!({ "creator": alice() }),
+            to_raw_json_value(&json!({ "creator": alice() })).unwrap(),
             &[],
             &[],
         );
-        let cre = create_event.event_id().clone();
+        let cre = create_event.event_id().to_owned();
         self.0.insert(cre.clone(), Arc::clone(&create_event));
 
         let alice_mem = to_pdu_event(
@@ -250,18 +257,18 @@ impl TestStore<StateEvent> {
             &[cre.clone()],
             &[cre.clone()],
         );
-        self.0.insert(alice_mem.event_id().clone(), Arc::clone(&alice_mem));
+        self.0.insert(alice_mem.event_id().to_owned(), Arc::clone(&alice_mem));
 
         let join_rules = to_pdu_event(
             "IJR",
             alice(),
             EventType::RoomJoinRules,
             Some(""),
-            json!({ "join_rule": JoinRule::Public }),
-            &[cre.clone(), alice_mem.event_id().clone()],
-            &[alice_mem.event_id().clone()],
+            to_raw_json_value(&RoomJoinRulesEventContent::new(JoinRule::Public)).unwrap(),
+            &[cre.clone(), alice_mem.event_id().to_owned()],
+            &[alice_mem.event_id().to_owned()],
         );
-        self.0.insert(join_rules.event_id().clone(), join_rules.clone());
+        self.0.insert(join_rules.event_id().to_owned(), join_rules.clone());
 
         // Bob and Charlie join at the same time, so there is a fork
         // this will be represented in the state_sets when we resolve
@@ -271,10 +278,10 @@ impl TestStore<StateEvent> {
             EventType::RoomMember,
             Some(bob().to_string().as_str()),
             member_content_join(),
-            &[cre.clone(), join_rules.event_id().clone()],
-            &[join_rules.event_id().clone()],
+            &[cre.clone(), join_rules.event_id().to_owned()],
+            &[join_rules.event_id().to_owned()],
         );
-        self.0.insert(bob_mem.event_id().clone(), bob_mem.clone());
+        self.0.insert(bob_mem.event_id().to_owned(), bob_mem.clone());
 
         let charlie_mem = to_pdu_event(
             "IMC",
@@ -282,68 +289,86 @@ impl TestStore<StateEvent> {
             EventType::RoomMember,
             Some(charlie().to_string().as_str()),
             member_content_join(),
-            &[cre, join_rules.event_id().clone()],
-            &[join_rules.event_id().clone()],
+            &[cre, join_rules.event_id().to_owned()],
+            &[join_rules.event_id().to_owned()],
         );
-        self.0.insert(charlie_mem.event_id().clone(), charlie_mem.clone());
+        self.0.insert(charlie_mem.event_id().to_owned(), charlie_mem.clone());
 
         let state_at_bob = [&create_event, &alice_mem, &join_rules, &bob_mem]
             .iter()
-            .map(|e| ((e.kind(), e.state_key().unwrap()), e.event_id().clone()))
+            .map(|e| {
+                (
+                    (e.event_type().to_owned(), e.state_key().unwrap().to_owned()),
+                    e.event_id().to_owned(),
+                )
+            })
             .collect::<StateMap<_>>();
 
         let state_at_charlie = [&create_event, &alice_mem, &join_rules, &charlie_mem]
             .iter()
-            .map(|e| ((e.kind(), e.state_key().unwrap()), e.event_id().clone()))
+            .map(|e| {
+                (
+                    (e.event_type().to_owned(), e.state_key().unwrap().to_owned()),
+                    e.event_id().to_owned(),
+                )
+            })
             .collect::<StateMap<_>>();
 
         let expected = [&create_event, &alice_mem, &join_rules, &bob_mem, &charlie_mem]
             .iter()
-            .map(|e| ((e.kind(), e.state_key().unwrap()), e.event_id().clone()))
+            .map(|e| {
+                (
+                    (e.event_type().to_owned(), e.state_key().unwrap().to_owned()),
+                    e.event_id().to_owned(),
+                )
+            })
             .collect::<StateMap<_>>();
 
         (state_at_bob, state_at_charlie, expected)
     }
 }
 
-fn event_id(id: &str) -> EventId {
+fn event_id(id: &str) -> Box<EventId> {
     if id.contains('$') {
-        return EventId::try_from(id).unwrap();
+        return id.try_into().unwrap();
     }
-    EventId::try_from(format!("${}:foo", id)).unwrap()
+    format!("${}:foo", id).try_into().unwrap()
 }
 
-fn alice() -> UserId {
-    UserId::try_from("@alice:foo").unwrap()
-}
-fn bob() -> UserId {
-    UserId::try_from("@bob:foo").unwrap()
-}
-fn charlie() -> UserId {
-    UserId::try_from("@charlie:foo").unwrap()
-}
-fn ella() -> UserId {
-    UserId::try_from("@ella:foo").unwrap()
+fn alice() -> Box<UserId> {
+    user_id!("@alice:foo").to_owned()
 }
 
-fn room_id() -> RoomId {
-    RoomId::try_from("!test:foo").unwrap()
+fn bob() -> Box<UserId> {
+    user_id!("@bob:foo").to_owned()
 }
 
-fn member_content_ban() -> JsonValue {
-    serde_json::to_value(MemberEventContent::new(MembershipState::Ban)).unwrap()
+fn charlie() -> Box<UserId> {
+    user_id!("@charlie:foo").to_owned()
 }
 
-fn member_content_join() -> JsonValue {
-    serde_json::to_value(MemberEventContent::new(MembershipState::Join)).unwrap()
+fn ella() -> Box<UserId> {
+    user_id!("@ella:foo").to_owned()
 }
 
-pub fn to_pdu_event<S>(
+fn room_id() -> &'static RoomId {
+    room_id!("!test:foo")
+}
+
+fn member_content_ban() -> Box<RawJsonValue> {
+    to_raw_json_value(&RoomMemberEventContent::new(MembershipState::Ban)).unwrap()
+}
+
+fn member_content_join() -> Box<RawJsonValue> {
+    to_raw_json_value(&RoomMemberEventContent::new(MembershipState::Join)).unwrap()
+}
+
+fn to_pdu_event<S>(
     id: &str,
-    sender: UserId,
+    sender: Box<UserId>,
     ev_type: EventType,
     state_key: Option<&str>,
-    content: JsonValue,
+    content: Box<RawJsonValue>,
     auth_events: &[S],
     prev_events: &[S],
 ) -> Arc<StateEvent>
@@ -359,9 +384,9 @@ where
 
     let state_key = state_key.map(ToOwned::to_owned);
     Arc::new(StateEvent {
-        event_id: EventId::try_from(id).unwrap(),
+        event_id: id.try_into().unwrap(),
         rest: Pdu::RoomV3Pdu(RoomV3Pdu {
-            room_id: room_id(),
+            room_id: room_id().to_owned(),
             sender,
             origin_server_ts: MilliSecondsSinceUnixEpoch(ts.try_into().unwrap()),
             state_key,
@@ -369,7 +394,7 @@ where
             content,
             redacts: None,
             unsigned: btreemap! {},
-            #[cfg(not(feature = "unstable-pre-spec"))]
+            #[cfg(not(feature = "__ci"))]
             origin: "foo".into(),
             auth_events,
             prev_events,
@@ -382,14 +407,14 @@ where
 
 // all graphs start with these input events
 #[allow(non_snake_case)]
-fn INITIAL_EVENTS() -> HashMap<EventId, Arc<StateEvent>> {
+fn INITIAL_EVENTS() -> HashMap<Box<EventId>, Arc<StateEvent>> {
     vec![
-        to_pdu_event::<EventId>(
+        to_pdu_event::<&EventId>(
             "CREATE",
             alice(),
             EventType::RoomCreate,
             Some(""),
-            json!({ "creator": alice() }),
+            to_raw_json_value(&json!({ "creator": alice() })).unwrap(),
             &[],
             &[],
         ),
@@ -407,7 +432,7 @@ fn INITIAL_EVENTS() -> HashMap<EventId, Arc<StateEvent>> {
             alice(),
             EventType::RoomPowerLevels,
             Some(""),
-            json!({ "users": { alice().to_string(): 100 } }),
+            to_raw_json_value(&json!({ "users": { alice().to_string(): 100 } })).unwrap(),
             &["CREATE", "IMA"],
             &["IMA"],
         ),
@@ -416,7 +441,7 @@ fn INITIAL_EVENTS() -> HashMap<EventId, Arc<StateEvent>> {
             alice(),
             EventType::RoomJoinRules,
             Some(""),
-            json!({ "join_rule": JoinRule::Public }),
+            to_raw_json_value(&RoomJoinRulesEventContent::new(JoinRule::Public)).unwrap(),
             &["CREATE", "IMA", "IPOWER"],
             &["IPOWER"],
         ),
@@ -438,40 +463,40 @@ fn INITIAL_EVENTS() -> HashMap<EventId, Arc<StateEvent>> {
             &["CREATE", "IJR", "IPOWER"],
             &["IMB"],
         ),
-        to_pdu_event::<EventId>(
+        to_pdu_event::<&EventId>(
             "START",
             charlie(),
             EventType::RoomTopic,
             Some(""),
-            json!({}),
+            to_raw_json_value(&json!({})).unwrap(),
             &[],
             &[],
         ),
-        to_pdu_event::<EventId>(
+        to_pdu_event::<&EventId>(
             "END",
             charlie(),
             EventType::RoomTopic,
             Some(""),
-            json!({}),
+            to_raw_json_value(&json!({})).unwrap(),
             &[],
             &[],
         ),
     ]
     .into_iter()
-    .map(|ev| (ev.event_id().clone(), ev))
+    .map(|ev| (ev.event_id().to_owned(), ev))
     .collect()
 }
 
 // all graphs start with these input events
 #[allow(non_snake_case)]
-fn BAN_STATE_SET() -> HashMap<EventId, Arc<StateEvent>> {
+fn BAN_STATE_SET() -> HashMap<Box<EventId>, Arc<StateEvent>> {
     vec![
         to_pdu_event(
             "PA",
             alice(),
             EventType::RoomPowerLevels,
             Some(""),
-            json!({ "users": { alice(): 100, bob(): 50 } }),
+            to_raw_json_value(&json!({ "users": { alice(): 100, bob(): 50 } })).unwrap(),
             &["CREATE", "IMA", "IPOWER"], // auth_events
             &["START"],                   // prev_events
         ),
@@ -480,7 +505,7 @@ fn BAN_STATE_SET() -> HashMap<EventId, Arc<StateEvent>> {
             alice(),
             EventType::RoomPowerLevels,
             Some(""),
-            json!({ "users": { alice(): 100, bob(): 50 } }),
+            to_raw_json_value(&json!({ "users": { alice(): 100, bob(): 50 } })).unwrap(),
             &["CREATE", "IMA", "IPOWER"],
             &["END"],
         ),
@@ -504,159 +529,35 @@ fn BAN_STATE_SET() -> HashMap<EventId, Arc<StateEvent>> {
         ),
     ]
     .into_iter()
-    .map(|ev| (ev.event_id().clone(), ev))
+    .map(|ev| (ev.event_id().to_owned(), ev))
     .collect()
 }
 
-pub mod event {
-    use std::collections::BTreeMap;
-
-    use js_int::UInt;
+mod event {
     use ruma_common::MilliSecondsSinceUnixEpoch;
-    use ruma_events::{
-        pdu::{EventHash, Pdu},
-        room::member::MembershipState,
-        EventType,
-    };
-    use ruma_identifiers::{
-        EventId, RoomId, RoomVersionId, ServerName, ServerSigningKeyId, UserId,
-    };
-    use ruma_serde::CanonicalJsonObject;
+    use ruma_events::{pdu::Pdu, EventType};
+    use ruma_identifiers::{EventId, RoomId, UserId};
     use ruma_state_res::Event;
     use serde::{Deserialize, Serialize};
-    use serde_json::Value as JsonValue;
+    use serde_json::value::RawValue as RawJsonValue;
 
     impl Event for StateEvent {
-        fn event_id(&self) -> &EventId {
-            self.event_id()
-        }
+        type Id = Box<EventId>;
 
-        fn room_id(&self) -> &RoomId {
-            self.room_id()
-        }
-
-        fn sender(&self) -> &UserId {
-            self.sender()
-        }
-
-        fn kind(&self) -> EventType {
-            self.kind()
-        }
-
-        fn content(&self) -> serde_json::Value {
-            self.content()
-        }
-
-        fn origin_server_ts(&self) -> MilliSecondsSinceUnixEpoch {
-            *self.origin_server_ts()
-        }
-
-        fn state_key(&self) -> Option<String> {
-            self.state_key()
-        }
-
-        fn prev_events(&self) -> Vec<EventId> {
-            self.prev_event_ids()
-        }
-
-        fn depth(&self) -> &UInt {
-            self.depth()
-        }
-
-        fn auth_events(&self) -> Vec<EventId> {
-            self.auth_events()
-        }
-
-        fn redacts(&self) -> Option<&EventId> {
-            self.redacts()
-        }
-
-        fn hashes(&self) -> &EventHash {
-            self.hashes()
-        }
-
-        fn signatures(&self) -> BTreeMap<Box<ServerName>, BTreeMap<ServerSigningKeyId, String>> {
-            self.signatures()
-        }
-
-        fn unsigned(&self) -> &BTreeMap<String, JsonValue> {
-            self.unsigned()
-        }
-    }
-
-    #[derive(Clone, Debug, Deserialize, Serialize)]
-    pub struct StateEvent {
-        pub event_id: EventId,
-        #[serde(flatten)]
-        pub rest: Pdu,
-    }
-
-    impl StateEvent {
-        pub fn from_id_value(id: EventId, json: serde_json::Value) -> serde_json::Result<Self> {
-            Ok(Self { event_id: id, rest: Pdu::RoomV3Pdu(serde_json::from_value(json)?) })
-        }
-
-        pub fn from_id_canon_obj(
-            id: EventId,
-            json: CanonicalJsonObject,
-        ) -> serde_json::Result<Self> {
-            Ok(Self {
-                event_id: id,
-                // TODO: this is unfortunate (from_value(to_value(json)))...
-                rest: Pdu::RoomV3Pdu(serde_json::from_value(serde_json::to_value(json)?)?),
-            })
-        }
-
-        pub fn is_power_event(&self) -> bool {
-            match &self.rest {
-                Pdu::RoomV1Pdu(event) => match event.kind {
-                    EventType::RoomPowerLevels
-                    | EventType::RoomJoinRules
-                    | EventType::RoomCreate => event.state_key == Some("".into()),
-                    EventType::RoomMember => {
-                        // TODO fix clone
-                        if let Ok(membership) = serde_json::from_value::<MembershipState>(
-                            event.content["membership"].clone(),
-                        ) {
-                            [MembershipState::Leave, MembershipState::Ban].contains(&membership)
-                                && event.sender.as_str()
-                                    // TODO is None here a failure
-                                    != event.state_key.as_deref().unwrap_or("NOT A STATE KEY")
-                        } else {
-                            false
-                        }
-                    }
-                    _ => false,
-                },
-                Pdu::RoomV3Pdu(event) => event.state_key == Some("".into()),
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn deserialize_content<C: serde::de::DeserializeOwned>(&self) -> serde_json::Result<C> {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => serde_json::from_value(ev.content.clone()),
-                Pdu::RoomV3Pdu(ev) => serde_json::from_value(ev.content.clone()),
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn origin_server_ts(&self) -> &MilliSecondsSinceUnixEpoch {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => &ev.origin_server_ts,
-                Pdu::RoomV3Pdu(ev) => &ev.origin_server_ts,
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn event_id(&self) -> &EventId {
+        fn event_id(&self) -> &Self::Id {
             &self.event_id
         }
 
-        pub fn sender(&self) -> &UserId {
+        fn room_id(&self) -> &RoomId {
+            match &self.rest {
+                Pdu::RoomV1Pdu(ev) => &ev.room_id,
+                Pdu::RoomV3Pdu(ev) => &ev.room_id,
+                #[cfg(not(feature = "unstable-exhaustive-types"))]
+                _ => unreachable!("new PDU version"),
+            }
+        }
+
+        fn sender(&self) -> &UserId {
             match &self.rest {
                 Pdu::RoomV1Pdu(ev) => &ev.sender,
                 Pdu::RoomV3Pdu(ev) => &ev.sender,
@@ -665,7 +566,61 @@ pub mod event {
             }
         }
 
-        pub fn redacts(&self) -> Option<&EventId> {
+        fn event_type(&self) -> &EventType {
+            match &self.rest {
+                Pdu::RoomV1Pdu(ev) => &ev.kind,
+                Pdu::RoomV3Pdu(ev) => &ev.kind,
+                #[cfg(not(feature = "unstable-exhaustive-types"))]
+                _ => unreachable!("new PDU version"),
+            }
+        }
+
+        fn content(&self) -> &RawJsonValue {
+            match &self.rest {
+                Pdu::RoomV1Pdu(ev) => &ev.content,
+                Pdu::RoomV3Pdu(ev) => &ev.content,
+                #[cfg(not(feature = "unstable-exhaustive-types"))]
+                _ => unreachable!("new PDU version"),
+            }
+        }
+
+        fn origin_server_ts(&self) -> MilliSecondsSinceUnixEpoch {
+            match &self.rest {
+                Pdu::RoomV1Pdu(ev) => ev.origin_server_ts,
+                Pdu::RoomV3Pdu(ev) => ev.origin_server_ts,
+                #[cfg(not(feature = "unstable-exhaustive-types"))]
+                _ => unreachable!("new PDU version"),
+            }
+        }
+
+        fn state_key(&self) -> Option<&str> {
+            match &self.rest {
+                Pdu::RoomV1Pdu(ev) => ev.state_key.as_deref(),
+                Pdu::RoomV3Pdu(ev) => ev.state_key.as_deref(),
+                #[cfg(not(feature = "unstable-exhaustive-types"))]
+                _ => unreachable!("new PDU version"),
+            }
+        }
+
+        fn prev_events(&self) -> Box<dyn DoubleEndedIterator<Item = &Self::Id> + '_> {
+            match &self.rest {
+                Pdu::RoomV1Pdu(ev) => Box::new(ev.prev_events.iter().map(|(id, _)| id)),
+                Pdu::RoomV3Pdu(ev) => Box::new(ev.prev_events.iter()),
+                #[cfg(not(feature = "unstable-exhaustive-types"))]
+                _ => unreachable!("new PDU version"),
+            }
+        }
+
+        fn auth_events(&self) -> Box<dyn DoubleEndedIterator<Item = &Self::Id> + '_> {
+            match &self.rest {
+                Pdu::RoomV1Pdu(ev) => Box::new(ev.auth_events.iter().map(|(id, _)| id)),
+                Pdu::RoomV3Pdu(ev) => Box::new(ev.auth_events.iter()),
+                #[cfg(not(feature = "unstable-exhaustive-types"))]
+                _ => unreachable!("new PDU version"),
+            }
+        }
+
+        fn redacts(&self) -> Option<&Self::Id> {
             match &self.rest {
                 Pdu::RoomV1Pdu(ev) => ev.redacts.as_ref(),
                 Pdu::RoomV3Pdu(ev) => ev.redacts.as_ref(),
@@ -673,132 +628,12 @@ pub mod event {
                 _ => unreachable!("new PDU version"),
             }
         }
+    }
 
-        pub fn room_id(&self) -> &RoomId {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => &ev.room_id,
-                Pdu::RoomV3Pdu(ev) => &ev.room_id,
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-        pub fn kind(&self) -> EventType {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => ev.kind.clone(),
-                Pdu::RoomV3Pdu(ev) => ev.kind.clone(),
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-        pub fn state_key(&self) -> Option<String> {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => ev.state_key.clone(),
-                Pdu::RoomV3Pdu(ev) => ev.state_key.clone(),
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        #[cfg(not(feature = "unstable-pre-spec"))]
-        pub fn origin(&self) -> String {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => ev.origin.clone(),
-                Pdu::RoomV3Pdu(ev) => ev.origin.clone(),
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn prev_event_ids(&self) -> Vec<EventId> {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => ev.prev_events.iter().map(|(id, _)| id).cloned().collect(),
-                Pdu::RoomV3Pdu(ev) => ev.prev_events.clone(),
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn auth_events(&self) -> Vec<EventId> {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => ev.auth_events.iter().map(|(id, _)| id).cloned().collect(),
-                Pdu::RoomV3Pdu(ev) => ev.auth_events.to_vec(),
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn content(&self) -> serde_json::Value {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => ev.content.clone(),
-                Pdu::RoomV3Pdu(ev) => ev.content.clone(),
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn unsigned(&self) -> &BTreeMap<String, serde_json::Value> {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => &ev.unsigned,
-                Pdu::RoomV3Pdu(ev) => &ev.unsigned,
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn signatures(
-            &self,
-        ) -> BTreeMap<Box<ServerName>, BTreeMap<ServerSigningKeyId, String>> {
-            match &self.rest {
-                Pdu::RoomV1Pdu(_) => maplit::btreemap! {},
-                Pdu::RoomV3Pdu(ev) => ev.signatures.clone(),
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn hashes(&self) -> &EventHash {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => &ev.hashes,
-                Pdu::RoomV3Pdu(ev) => &ev.hashes,
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn depth(&self) -> &UInt {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => &ev.depth,
-                Pdu::RoomV3Pdu(ev) => &ev.depth,
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        pub fn is_type_and_key(&self, ev_type: EventType, state_key: &str) -> bool {
-            match &self.rest {
-                Pdu::RoomV1Pdu(ev) => {
-                    ev.kind == ev_type && ev.state_key.as_deref() == Some(state_key)
-                }
-                Pdu::RoomV3Pdu(ev) => {
-                    ev.kind == ev_type && ev.state_key.as_deref() == Some(state_key)
-                }
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
-
-        /// Returns the room version this event is formatted for.
-        ///
-        /// Currently either version 1 or 6 is returned, 6 represents
-        /// version 3 and above.
-        pub fn room_version(&self) -> RoomVersionId {
-            // TODO: We have to know the actual room version this is not sufficient
-            match self.rest {
-                Pdu::RoomV1Pdu(_) => RoomVersionId::Version1,
-                Pdu::RoomV3Pdu(_) => RoomVersionId::Version6,
-                #[cfg(not(feature = "unstable-exhaustive-types"))]
-                _ => unreachable!("new PDU version"),
-            }
-        }
+    #[derive(Clone, Debug, Deserialize, Serialize)]
+    pub struct StateEvent {
+        pub event_id: Box<EventId>,
+        #[serde(flatten)]
+        pub rest: Pdu,
     }
 }
